@@ -19,6 +19,10 @@
 
 require 'mkmf'
 
+# set the ruby version compiler flag
+runtime_version = RUBY_VERSION.gsub(/\./,'')[0,2]
+$CFLAGS << " -DRUBY#{runtime_version}"
+
 dir_config("qpid-proton")
 
 REQUIRED_LIBRARIES = [
@@ -33,7 +37,6 @@ REQUIRED_HEADERS = [
                     "proton/engine.h",
                     "proton/message.h",
                     "proton/sasl.h",
-                    "proton/driver.h",
                     "proton/messenger.h"
                    ]
 
