@@ -22,7 +22,6 @@
 #include <proton/object.h>
 #include <proton/codec.h>
 #include <proton/error.h>
-#include <proton/util.h>
 #include <assert.h>
 #include <stdio.h>
 #include <string.h>
@@ -32,12 +31,13 @@
 #include "encodings.h"
 #define DEFINE_FIELDS
 #include "protocol.h"
-#include "../platform.h"
-#include "../platform_fmt.h"
-#include "../util.h"
+#include "platform.h"
+#include "platform_fmt.h"
+#include "util.h"
 #include "decoder.h"
 #include "encoder.h"
 #include "data.h"
+#include "log_private.h"
 
 const char *pn_type_name(pn_type_t type)
 {
@@ -92,7 +92,7 @@ static void pn_data_finalize(void *object)
   pn_free(data->encoder);
 }
 
-static pn_fields_t *pni_node_fields(pn_data_t *data, pni_node_t *node)
+static const pn_fields_t *pni_node_fields(pn_data_t *data, pni_node_t *node)
 {
   if (!node) return NULL;
   if (node->atom.type != PN_DESCRIBED) return NULL;
@@ -103,8 +103,9 @@ static pn_fields_t *pni_node_fields(pn_data_t *data, pni_node_t *node)
     return NULL;
   }
 
-  if (descriptor->atom.u.as_ulong < 256) {
-    return &FIELDS[descriptor->atom.u.as_ulong];
+  if (descriptor->atom.u.as_ulong >= FIELD_MIN && descriptor->atom.u.as_ulong <= FIELD_MAX) {
+    const pn_fields_t *f = &FIELDS[descriptor->atom.u.as_ulong-FIELD_MIN];
+    return (f->name_index!=0) ? f : NULL;
   } else {
     return NULL;
   }
@@ -140,7 +141,7 @@ int pni_inspect_atom(pn_atom_t *atom, pn_string_t *str)
   case PN_INT:
     return pn_string_addf(str, "%" PRIi32, atom->u.as_int);
   case PN_CHAR:
-    return pn_string_addf(str, "%lc",  atom->u.as_char);
+    return pn_string_addf(str, "%c",  atom->u.as_char);
   case PN_ULONG:
     return pn_string_addf(str, "%" PRIu64, atom->u.as_ulong);
   case PN_LONG:
@@ -233,9 +234,16 @@ int pni_inspect_atom(pn_atom_t *atom, pn_string_t *str)
       if (quote) if ((err = pn_string_addf(str, "\""))) return err;
       return 0;
     }
+  case PN_LIST:
+    return pn_string_addf(str, "<list>");
+  case PN_MAP:
+    return pn_string_addf(str, "<map>");
+  case PN_ARRAY:
+    return pn_string_addf(str, "<array>");
+  case PN_DESCRIBED:
+    return pn_string_addf(str, "<described>");
   default:
-    assert(false);
-    return PN_ERR;
+    return pn_string_addf(str, "<undefined: %i>", atom->type);
   }
 }
 
@@ -245,9 +253,9 @@ int pni_inspect_enter(void *ctx, pn_data_t *data, pni_node_t *node)
   pn_atom_t *atom = (pn_atom_t *) &node->atom;
 
   pni_node_t *parent = pn_data_node(data, node->parent);
-  pn_fields_t *fields = pni_node_fields(data, parent);
+  const pn_fields_t *fields = pni_node_fields(data, parent);
   pni_node_t *grandparent = parent ? pn_data_node(data, parent->parent) : NULL;
-  pn_fields_t *grandfields = pni_node_fields(data, grandparent);
+  const pn_fields_t *grandfields = pni_node_fields(data, grandparent);
   int index = pni_node_index(data, node);
 
   int err;
@@ -256,7 +264,9 @@ int pni_inspect_enter(void *ctx, pn_data_t *data, pni_node_t *node)
     if (atom->type == PN_NULL) {
       return 0;
     }
-    const char *name = grandfields->fields[index];
+    const char *name = (index < grandfields->field_count)
+        ? FIELD_STRINGPOOL+FIELD_FIELDS[grandfields->first_field_index+index]
+        : NULL;
     if (name) {
       err = pn_string_addf(str, "%s=", name);
       if (err) return err;
@@ -275,7 +285,7 @@ int pni_inspect_enter(void *ctx, pn_data_t *data, pni_node_t *node)
     return pn_string_addf(str, "{");
   default:
     if (fields && index == 0) {
-      err = pn_string_addf(str, "%s", fields->name);
+      err = pn_string_addf(str, "%s", FIELD_STRINGPOOL+FIELD_NAME[fields->name_index]);
       if (err) return err;
       err = pn_string_addf(str, "(");
       if (err) return err;
@@ -305,7 +315,7 @@ int pni_inspect_exit(void *ctx, pn_data_t *data, pni_node_t *node)
   pn_string_t *str = (pn_string_t *) ctx;
   pni_node_t *parent = pn_data_node(data, node->parent);
   pni_node_t *grandparent = parent ? pn_data_node(data, parent->parent) : NULL;
-  pn_fields_t *grandfields = pni_node_fields(data, grandparent);
+  const pn_fields_t *grandfields = pni_node_fields(data, grandparent);
   pni_node_t *next = pn_data_node(data, node->next);
   int err;
 
@@ -356,8 +366,8 @@ static int pn_data_inspect(void *obj, pn_string_t *dst)
 
 pn_data_t *pn_data(size_t capacity)
 {
-  static pn_class_t clazz = PN_CLASS(pn_data);
-  pn_data_t *data = (pn_data_t *) pn_new(sizeof(pn_data_t), &clazz);
+  static const pn_class_t clazz = PN_CLASS(pn_data);
+  pn_data_t *data = (pn_data_t *) pn_class_new(&clazz, sizeof(pn_data_t));
   data->capacity = capacity;
   data->size = 0;
   data->nodes = capacity ? (pni_node_t *) malloc(capacity * sizeof(pni_node_t)) : NULL;
@@ -405,14 +415,17 @@ void pn_data_clear(pn_data_t *data)
   }
 }
 
-int pn_data_grow(pn_data_t *data)
+static int pni_data_grow(pn_data_t *data)
 {
-  data->capacity = 2*(data->capacity ? data->capacity : 16);
-  data->nodes = (pni_node_t *) realloc(data->nodes, data->capacity * sizeof(pni_node_t));
+  pni_nid_t new_capacity = 2 * (data->capacity ? data->capacity : 2);
+  pni_node_t *new_nodes = (pni_node_t *)realloc(data->nodes, new_capacity * sizeof(pni_node_t));
+  if (new_nodes == NULL) return PN_OUT_OF_MEMORY;
+  data->capacity = new_capacity;
+  data->nodes = new_nodes;
   return 0;
 }
 
-ssize_t pn_data_intern(pn_data_t *data, char *start, size_t size)
+static ssize_t pni_data_intern(pn_data_t *data, const char *start, size_t size)
 {
   size_t offset = pn_buffer_size(data->buf);
   int err = pn_buffer_append(data->buf, start, size);
@@ -422,7 +435,7 @@ ssize_t pn_data_intern(pn_data_t *data, char *start, size_t size)
   return offset;
 }
 
-pn_bytes_t *pn_data_bytes(pn_data_t *data, pni_node_t *node)
+static pn_bytes_t *pni_data_bytes(pn_data_t *data, pni_node_t *node)
 {
   switch (node->atom.type) {
   case PN_BINARY:
@@ -433,32 +446,32 @@ pn_bytes_t *pn_data_bytes(pn_data_t *data, pni_node_t *node)
   }
 }
 
-void pn_data_rebase(pn_data_t *data, char *base)
+static void pni_data_rebase(pn_data_t *data, char *base)
 {
   for (unsigned i = 0; i < data->size; i++) {
     pni_node_t *node = &data->nodes[i];
     if (node->data) {
-      pn_bytes_t *bytes = pn_data_bytes(data, node);
+      pn_bytes_t *bytes = pni_data_bytes(data, node);
       bytes->start = base + node->data_offset;
     }
   }
 }
 
-int pn_data_intern_node(pn_data_t *data, pni_node_t *node)
+static int pni_data_intern_node(pn_data_t *data, pni_node_t *node)
 {
-  pn_bytes_t *bytes = pn_data_bytes(data, node);
+  pn_bytes_t *bytes = pni_data_bytes(data, node);
   if (!bytes) return 0;
   size_t oldcap = pn_buffer_capacity(data->buf);
-  ssize_t offset = pn_data_intern(data, bytes->start, bytes->size);
+  ssize_t offset = pni_data_intern(data, bytes->start, bytes->size);
   if (offset < 0) return offset;
   node->data = true;
   node->data_offset = offset;
   node->data_size = bytes->size;
-  pn_bytes_t buf = pn_buffer_bytes(data->buf);
+  pn_buffer_memory_t buf = pn_buffer_memory(data->buf);
   bytes->start = buf.start + offset;
 
   if (pn_buffer_capacity(data->buf) != oldcap) {
-    pn_data_rebase(data, buf.start);
+    pni_data_rebase(data, buf.start);
   }
 
   return 0;
@@ -466,7 +479,8 @@ int pn_data_intern_node(pn_data_t *data, pni_node_t *node)
 
 int pn_data_vfill(pn_data_t *data, const char *fmt, va_list ap)
 {
-  int err;
+  int err = 0;
+  const char *begin = fmt;
   while (*fmt) {
     char code = *(fmt++);
     if (!code) return 0;
@@ -568,7 +582,7 @@ int pn_data_vfill(pn_data_t *data, const char *fmt, va_list ap)
       }
       break;
     case '[':
-      if (*(fmt - 2) != 'T') {
+      if (fmt < (begin + 2) || *(fmt - 2) != 'T') {
         err = pn_data_put_list(data);
         if (err) return err;
         pn_data_enter(data);
@@ -612,7 +626,7 @@ int pn_data_vfill(pn_data_t *data, const char *fmt, va_list ap)
           }
           break;
         default:
-          fprintf(stderr, "unrecognized * code: 0x%.2X '%c'\n", code, code);
+          pn_logf("unrecognized * code: 0x%.2X '%c'", code, code);
           return PN_ARG_ERR;
         }
       }
@@ -630,7 +644,7 @@ int pn_data_vfill(pn_data_t *data, const char *fmt, va_list ap)
       }
       break;
     default:
-      fprintf(stderr, "unrecognized fill code: 0x%.2X '%c'\n", code, code);
+      pn_logf("unrecognized fill code: 0x%.2X '%c'", code, code);
       return PN_ARG_ERR;
     }
 
@@ -646,7 +660,6 @@ int pn_data_vfill(pn_data_t *data, const char *fmt, va_list ap)
         pni_node_t *current = pn_data_node(data, data->current);
         current->down = 0;
         current->children = 0;
-        if (err) return err;
         parent = pn_data_node(data, data->parent);
       } else {
         break;
@@ -686,7 +699,7 @@ static bool pn_scan_next(pn_data_t *data, pn_type_t *type, bool suspend)
   }
 }
 
-pni_node_t *pn_data_peek(pn_data_t *data);
+static pni_node_t *pni_data_peek(pn_data_t *data);
 
 int pn_data_vscan(pn_data_t *data, const char *fmt, va_list ap)
 {
@@ -1027,7 +1040,7 @@ int pn_data_vscan(pn_data_t *data, const char *fmt, va_list ap)
         pn_data_t *dst = va_arg(ap, pn_data_t *);
         if (!suspend) {
           size_t old = pn_data_size(dst);
-          pni_node_t *next = pn_data_peek(data);
+          pni_node_t *next = pni_data_peek(data);
           if (next && next->atom.type != PN_NULL) {
             pn_data_narrow(data);
             int err = pn_data_appendn(dst, data, 1);
@@ -1094,33 +1107,14 @@ int pn_data_format(pn_data_t *data, char *bytes, size_t *size)
   }
 }
 
-int pn_data_resize(pn_data_t *data, size_t size)
-{
-  if (!data || size > data->capacity) return PN_ARG_ERR;
-  data->size = size;
-  return 0;
-}
-
-
-pni_node_t *pn_data_node(pn_data_t *data, size_t nd)
-{
-  if (nd) {
-    return &data->nodes[nd - 1];
-  } else {
-    return NULL;
-  }
-}
-
-size_t pn_data_id(pn_data_t *data, pni_node_t *node)
+static size_t pni_data_id(pn_data_t *data, pni_node_t *node)
 {
   return node - data->nodes + 1;
 }
 
-pni_node_t *pn_data_new(pn_data_t *data)
+static pni_node_t *pni_data_new(pn_data_t *data)
 {
-  if (data->capacity <= data->size) {
-    pn_data_grow(data);
-  }
+  if ((data->capacity <= data->size) && (pni_data_grow(data) != 0)) return NULL;
   pni_node_t *node = pn_data_node(data, ++(data->size));
   node->next = 0;
   node->down = 0;
@@ -1134,7 +1128,7 @@ void pn_data_rewind(pn_data_t *data)
   data->current = data->base_current;
 }
 
-pni_node_t *pn_data_current(pn_data_t *data)
+static pni_node_t *pni_data_current(pn_data_t *data)
 {
   return pn_data_node(data, data->current);
 }
@@ -1163,13 +1157,13 @@ pn_handle_t pn_data_point(pn_data_t *data)
 bool pn_data_restore(pn_data_t *data, pn_handle_t point)
 {
   pn_shandle_t spoint = (pn_shandle_t) point;
-  if (spoint < 0 && ((size_t) (-spoint)) <= data->size) {
+  if (spoint <= 0 && ((size_t) (-spoint)) <= data->size) {
     data->parent = -((pn_shandle_t) point);
     data->current = 0;
     return true;
   } else if (point && point <= data->size) {
     data->current = point;
-    pni_node_t *current = pn_data_current(data);
+    pni_node_t *current = pni_data_current(data);
     data->parent = current->parent;
     return true;
   } else {
@@ -1177,9 +1171,9 @@ bool pn_data_restore(pn_data_t *data, pn_handle_t point)
   }
 }
 
-pni_node_t *pn_data_peek(pn_data_t *data)
+static pni_node_t *pni_data_peek(pn_data_t *data)
 {
-  pni_node_t *current = pn_data_current(data);
+  pni_node_t *current = pni_data_current(data);
   if (current) {
     return pn_data_node(data, current->next);
   }
@@ -1194,7 +1188,7 @@ pni_node_t *pn_data_peek(pn_data_t *data)
 
 bool pn_data_next(pn_data_t *data)
 {
-  pni_node_t *current = pn_data_current(data);
+  pni_node_t *current = pni_data_current(data);
   pni_node_t *parent = pn_data_node(data, data->parent);
   size_t next;
 
@@ -1218,7 +1212,7 @@ bool pn_data_next(pn_data_t *data)
 
 bool pn_data_prev(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->prev) {
     data->current = node->prev;
     return true;
@@ -1269,7 +1263,7 @@ int pni_data_traverse(pn_data_t *data,
 
 pn_type_t pn_data_type(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node) {
     return node->atom.type;
   } else {
@@ -1330,7 +1324,7 @@ bool pn_data_lookup(pn_data_t *data, const char *name)
     case PN_SYMBOL:
       {
         pn_bytes_t bytes = pn_data_get_bytes(data);
-        if (!strncmp(name, bytes.start, bytes.size)) {
+        if (strlen(name) == bytes.size && !strncmp(name, bytes.start, bytes.size)) {
           return pn_data_next(data);
         }
       }
@@ -1348,7 +1342,7 @@ bool pn_data_lookup(pn_data_t *data, const char *name)
 
 void pn_data_dump(pn_data_t *data)
 {
-  printf("{current=%" PN_ZI ", parent=%" PN_ZI "}\n", data->current, data->parent);
+  printf("{current=%" PN_ZI ", parent=%" PN_ZI "}\n", (size_t) data->current, (size_t) data->parent);
   for (unsigned i = 0; i < data->size; i++)
   {
     pni_node_t *node = &data->nodes[i];
@@ -1356,14 +1350,18 @@ void pn_data_dump(pn_data_t *data)
     pni_inspect_atom((pn_atom_t *) &node->atom, data->str);
     printf("Node %i: prev=%" PN_ZI ", next=%" PN_ZI ", parent=%" PN_ZI ", down=%" PN_ZI 
            ", children=%" PN_ZI ", type=%s (%s)\n",
-           i + 1, node->prev, node->next, node->parent, node->down, node->children,
+           i + 1, (size_t) node->prev,
+           (size_t) node->next,
+           (size_t) node->parent,
+           (size_t) node->down,
+           (size_t) node->children,
            pn_type_name(node->atom.type), pn_string_get(data->str));
   }
 }
 
-pni_node_t *pn_data_add(pn_data_t *data)
+static pni_node_t *pni_data_add(pn_data_t *data)
 {
-  pni_node_t *current = pn_data_current(data);
+  pni_node_t *current = pni_data_current(data);
   pni_node_t *parent = pn_data_node(data, data->parent);
   pni_node_t *node;
 
@@ -1371,16 +1369,18 @@ pni_node_t *pn_data_add(pn_data_t *data)
     if (current->next) {
       node = pn_data_node(data, current->next);
     } else {
-      node = pn_data_new(data);
+      node = pni_data_new(data);
+      if (!node) return NULL;
+
       // refresh the pointers in case we grew
-      current = pn_data_current(data);
+      current = pni_data_current(data);
       parent = pn_data_node(data, data->parent);
       node->prev = data->current;
-      current->next = pn_data_id(data, node);
+      current->next = pni_data_id(data, node);
       node->parent = data->parent;
       if (parent) {
         if (!parent->down) {
-          parent->down = pn_data_id(data, node);
+          parent->down = pni_data_id(data, node);
         }
         parent->children++;
       }
@@ -1389,18 +1389,22 @@ pni_node_t *pn_data_add(pn_data_t *data)
     if (parent->down) {
       node = pn_data_node(data, parent->down);
     } else {
-      node = pn_data_new(data);
+      node = pni_data_new(data);
+      if (!node) return NULL;
+
       // refresh the pointers in case we grew
       parent = pn_data_node(data, data->parent);
       node->prev = 0;
       node->parent = data->parent;
-      parent->down = pn_data_id(data, node);
+      parent->down = pni_data_id(data, node);
       parent->children++;
     }
   } else if (data->size) {
     node = pn_data_node(data, 1);
   } else {
-    node = pn_data_new(data);
+    node = pni_data_new(data);
+    if (!node) return NULL;
+
     node->prev = 0;
     node->parent = 0;
   }
@@ -1410,7 +1414,7 @@ pni_node_t *pn_data_add(pn_data_t *data)
   node->data = false;
   node->data_offset = 0;
   node->data_size = 0;
-  data->current = pn_data_id(data, node);
+  data->current = pni_data_id(data, node);
   return node;
 }
 
@@ -1419,6 +1423,11 @@ ssize_t pn_data_encode(pn_data_t *data, char *bytes, size_t size)
   return pn_encoder_encode(data->encoder, data, bytes, size);
 }
 
+ssize_t pn_data_encoded_size(pn_data_t *data)
+{
+  return pn_encoder_size(data->encoder, data);
+}
+
 ssize_t pn_data_decode(pn_data_t *data, const char *bytes, size_t size)
 {
   return pn_decoder_decode(data->decoder, bytes, size, data);
@@ -1426,21 +1435,24 @@ ssize_t pn_data_decode(pn_data_t *data, const char *bytes, size_t size)
 
 int pn_data_put_list(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_LIST;
   return 0;
 }
 
 int pn_data_put_map(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_MAP;
   return 0;
 }
 
 int pn_data_put_array(pn_data_t *data, bool described, pn_type_t type)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_ARRAY;
   node->described = described;
   node->type = type;
@@ -1449,27 +1461,30 @@ int pn_data_put_array(pn_data_t *data, bool described, pn_type_t type)
 
 void pni_data_set_array_type(pn_data_t *data, pn_type_t type)
 {
-  pni_node_t *array = pn_data_current(data);
-  array->type = type;
+  pni_node_t *array = pni_data_current(data);
+  if (array) array->type = type;
 }
 
 int pn_data_put_described(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_DESCRIBED;
   return 0;
 }
 
 int pn_data_put_null(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   pni_atom_init(&node->atom, PN_NULL);
   return 0;
 }
 
 int pn_data_put_bool(pn_data_t *data, bool b)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_BOOL;
   node->atom.u.as_bool = b;
   return 0;
@@ -1477,7 +1492,8 @@ int pn_data_put_bool(pn_data_t *data, bool b)
 
 int pn_data_put_ubyte(pn_data_t *data, uint8_t ub)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_UBYTE;
   node->atom.u.as_ubyte = ub;
   return 0;
@@ -1485,7 +1501,8 @@ int pn_data_put_ubyte(pn_data_t *data, uint8_t ub)
 
 int pn_data_put_byte(pn_data_t *data, int8_t b)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_BYTE;
   node->atom.u.as_byte = b;
   return 0;
@@ -1493,7 +1510,8 @@ int pn_data_put_byte(pn_data_t *data, int8_t b)
 
 int pn_data_put_ushort(pn_data_t *data, uint16_t us)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_USHORT;
   node->atom.u.as_ushort = us;
   return 0;
@@ -1501,7 +1519,8 @@ int pn_data_put_ushort(pn_data_t *data, uint16_t us)
 
 int pn_data_put_short(pn_data_t *data, int16_t s)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_SHORT;
   node->atom.u.as_short = s;
   return 0;
@@ -1509,7 +1528,8 @@ int pn_data_put_short(pn_data_t *data, int16_t s)
 
 int pn_data_put_uint(pn_data_t *data, uint32_t ui)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_UINT;
   node->atom.u.as_uint = ui;
   return 0;
@@ -1517,7 +1537,8 @@ int pn_data_put_uint(pn_data_t *data, uint32_t ui)
 
 int pn_data_put_int(pn_data_t *data, int32_t i)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_INT;
   node->atom.u.as_int = i;
   return 0;
@@ -1525,7 +1546,8 @@ int pn_data_put_int(pn_data_t *data, int32_t i)
 
 int pn_data_put_char(pn_data_t *data, pn_char_t c)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_CHAR;
   node->atom.u.as_char = c;
   return 0;
@@ -1533,7 +1555,8 @@ int pn_data_put_char(pn_data_t *data, pn_char_t c)
 
 int pn_data_put_ulong(pn_data_t *data, uint64_t ul)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_ULONG;
   node->atom.u.as_ulong = ul;
   return 0;
@@ -1541,7 +1564,8 @@ int pn_data_put_ulong(pn_data_t *data, uint64_t ul)
 
 int pn_data_put_long(pn_data_t *data, int64_t l)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_LONG;
   node->atom.u.as_long = l;
   return 0;
@@ -1549,7 +1573,8 @@ int pn_data_put_long(pn_data_t *data, int64_t l)
 
 int pn_data_put_timestamp(pn_data_t *data, pn_timestamp_t t)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_TIMESTAMP;
   node->atom.u.as_timestamp = t;
   return 0;
@@ -1557,7 +1582,8 @@ int pn_data_put_timestamp(pn_data_t *data, pn_timestamp_t t)
 
 int pn_data_put_float(pn_data_t *data, float f)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_FLOAT;
   node->atom.u.as_float = f;
   return 0;
@@ -1565,7 +1591,8 @@ int pn_data_put_float(pn_data_t *data, float f)
 
 int pn_data_put_double(pn_data_t *data, double d)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_DOUBLE;
   node->atom.u.as_double = d;
   return 0;
@@ -1573,7 +1600,8 @@ int pn_data_put_double(pn_data_t *data, double d)
 
 int pn_data_put_decimal32(pn_data_t *data, pn_decimal32_t d)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_DECIMAL32;
   node->atom.u.as_decimal32 = d;
   return 0;
@@ -1581,7 +1609,8 @@ int pn_data_put_decimal32(pn_data_t *data, pn_decimal32_t d)
 
 int pn_data_put_decimal64(pn_data_t *data, pn_decimal64_t d)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_DECIMAL64;
   node->atom.u.as_decimal64 = d;
   return 0;
@@ -1589,7 +1618,8 @@ int pn_data_put_decimal64(pn_data_t *data, pn_decimal64_t d)
 
 int pn_data_put_decimal128(pn_data_t *data, pn_decimal128_t d)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_DECIMAL128;
   memmove(node->atom.u.as_decimal128.bytes, d.bytes, 16);
   return 0;
@@ -1597,7 +1627,8 @@ int pn_data_put_decimal128(pn_data_t *data, pn_decimal128_t d)
 
 int pn_data_put_uuid(pn_data_t *data, pn_uuid_t u)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_UUID;
   memmove(node->atom.u.as_uuid.bytes, u.bytes, 16);
   return 0;
@@ -1605,38 +1636,42 @@ int pn_data_put_uuid(pn_data_t *data, pn_uuid_t u)
 
 int pn_data_put_binary(pn_data_t *data, pn_bytes_t bytes)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_BINARY;
   node->atom.u.as_bytes = bytes;
-  return pn_data_intern_node(data, node);
+  return pni_data_intern_node(data, node);
 }
 
 int pn_data_put_string(pn_data_t *data, pn_bytes_t string)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_STRING;
   node->atom.u.as_bytes = string;
-  return pn_data_intern_node(data, node);
+  return pni_data_intern_node(data, node);
 }
 
 int pn_data_put_symbol(pn_data_t *data, pn_bytes_t symbol)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom.type = PN_SYMBOL;
   node->atom.u.as_bytes = symbol;
-  return pn_data_intern_node(data, node);
+  return pni_data_intern_node(data, node);
 }
 
 int pn_data_put_atom(pn_data_t *data, pn_atom_t atom)
 {
-  pni_node_t *node = pn_data_add(data);
+  pni_node_t *node = pni_data_add(data);
+  if (node == NULL) return PN_OUT_OF_MEMORY;
   node->atom = atom;
-  return pn_data_intern_node(data, node);
+  return pni_data_intern_node(data, node);
 }
 
 size_t pn_data_get_list(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_LIST) {
     return node->children;
   } else {
@@ -1646,7 +1681,7 @@ size_t pn_data_get_list(pn_data_t *data)
 
 size_t pn_data_get_map(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_MAP) {
     return node->children;
   } else {
@@ -1656,7 +1691,7 @@ size_t pn_data_get_map(pn_data_t *data)
 
 size_t pn_data_get_array(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_ARRAY) {
     if (node->described) {
       return node->children - 1;
@@ -1670,7 +1705,7 @@ size_t pn_data_get_array(pn_data_t *data)
 
 bool pn_data_is_array_described(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_ARRAY) {
     return node->described;
   } else {
@@ -1680,7 +1715,7 @@ bool pn_data_is_array_described(pn_data_t *data)
 
 pn_type_t pn_data_get_array_type(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_ARRAY) {
     return node->type;
   } else {
@@ -1690,19 +1725,19 @@ pn_type_t pn_data_get_array_type(pn_data_t *data)
 
 bool pn_data_is_described(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   return node && node->atom.type == PN_DESCRIBED;
 }
 
 bool pn_data_is_null(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   return node && node->atom.type == PN_NULL;
 }
 
 bool pn_data_get_bool(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_BOOL) {
     return node->atom.u.as_bool;
   } else {
@@ -1712,7 +1747,7 @@ bool pn_data_get_bool(pn_data_t *data)
 
 uint8_t pn_data_get_ubyte(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_UBYTE) {
     return node->atom.u.as_ubyte;
   } else {
@@ -1722,7 +1757,7 @@ uint8_t pn_data_get_ubyte(pn_data_t *data)
 
 int8_t pn_data_get_byte(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_BYTE) {
     return node->atom.u.as_byte;
   } else {
@@ -1732,7 +1767,7 @@ int8_t pn_data_get_byte(pn_data_t *data)
 
 uint16_t pn_data_get_ushort(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_USHORT) {
     return node->atom.u.as_ushort;
   } else {
@@ -1742,7 +1777,7 @@ uint16_t pn_data_get_ushort(pn_data_t *data)
 
 int16_t pn_data_get_short(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_SHORT) {
     return node->atom.u.as_short;
   } else {
@@ -1752,7 +1787,7 @@ int16_t pn_data_get_short(pn_data_t *data)
 
 uint32_t pn_data_get_uint(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_UINT) {
     return node->atom.u.as_uint;
   } else {
@@ -1762,7 +1797,7 @@ uint32_t pn_data_get_uint(pn_data_t *data)
 
 int32_t pn_data_get_int(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_INT) {
     return node->atom.u.as_int;
   } else {
@@ -1772,7 +1807,7 @@ int32_t pn_data_get_int(pn_data_t *data)
 
 pn_char_t pn_data_get_char(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_CHAR) {
     return node->atom.u.as_char;
   } else {
@@ -1782,7 +1817,7 @@ pn_char_t pn_data_get_char(pn_data_t *data)
 
 uint64_t pn_data_get_ulong(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_ULONG) {
     return node->atom.u.as_ulong;
   } else {
@@ -1792,7 +1827,7 @@ uint64_t pn_data_get_ulong(pn_data_t *data)
 
 int64_t pn_data_get_long(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_LONG) {
     return node->atom.u.as_long;
   } else {
@@ -1802,7 +1837,7 @@ int64_t pn_data_get_long(pn_data_t *data)
 
 pn_timestamp_t pn_data_get_timestamp(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_TIMESTAMP) {
     return node->atom.u.as_timestamp;
   } else {
@@ -1812,7 +1847,7 @@ pn_timestamp_t pn_data_get_timestamp(pn_data_t *data)
 
 float pn_data_get_float(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_FLOAT) {
     return node->atom.u.as_float;
   } else {
@@ -1822,7 +1857,7 @@ float pn_data_get_float(pn_data_t *data)
 
 double pn_data_get_double(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_DOUBLE) {
     return node->atom.u.as_double;
   } else {
@@ -1832,7 +1867,7 @@ double pn_data_get_double(pn_data_t *data)
 
 pn_decimal32_t pn_data_get_decimal32(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_DECIMAL32) {
     return node->atom.u.as_decimal32;
   } else {
@@ -1842,7 +1877,7 @@ pn_decimal32_t pn_data_get_decimal32(pn_data_t *data)
 
 pn_decimal64_t pn_data_get_decimal64(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_DECIMAL64) {
     return node->atom.u.as_decimal64;
   } else {
@@ -1852,7 +1887,7 @@ pn_decimal64_t pn_data_get_decimal64(pn_data_t *data)
 
 pn_decimal128_t pn_data_get_decimal128(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_DECIMAL128) {
     return node->atom.u.as_decimal128;
   } else {
@@ -1863,7 +1898,7 @@ pn_decimal128_t pn_data_get_decimal128(pn_data_t *data)
 
 pn_uuid_t pn_data_get_uuid(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_UUID) {
     return node->atom.u.as_uuid;
   } else {
@@ -1874,7 +1909,7 @@ pn_uuid_t pn_data_get_uuid(pn_data_t *data)
 
 pn_bytes_t pn_data_get_binary(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_BINARY) {
     return node->atom.u.as_bytes;
   } else {
@@ -1885,7 +1920,7 @@ pn_bytes_t pn_data_get_binary(pn_data_t *data)
 
 pn_bytes_t pn_data_get_string(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_STRING) {
     return node->atom.u.as_bytes;
   } else {
@@ -1896,7 +1931,7 @@ pn_bytes_t pn_data_get_string(pn_data_t *data)
 
 pn_bytes_t pn_data_get_symbol(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && node->atom.type == PN_SYMBOL) {
     return node->atom.u.as_bytes;
   } else {
@@ -1907,7 +1942,7 @@ pn_bytes_t pn_data_get_symbol(pn_data_t *data)
 
 pn_bytes_t pn_data_get_bytes(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node && (node->atom.type == PN_BINARY ||
                node->atom.type == PN_STRING ||
                node->atom.type == PN_SYMBOL)) {
@@ -1920,11 +1955,11 @@ pn_bytes_t pn_data_get_bytes(pn_data_t *data)
 
 pn_atom_t pn_data_get_atom(pn_data_t *data)
 {
-  pni_node_t *node = pn_data_current(data);
+  pni_node_t *node = pni_data_current(data);
   if (node) {
     return *((pn_atom_t *) &node->atom);
   } else {
-    pn_atom_t t = {PN_NULL};
+    pn_atom_t t = {PN_NULL, {0,}};
     return t;
   }
 }
