@@ -23,9 +23,6 @@ package org.apache.qpid.proton.engine.impl;
 import org.apache.qpid.proton.engine.Collector;
 import org.apache.qpid.proton.engine.Event;
 
-import java.util.LinkedList;
-import java.util.Queue;
-
 
 /**
  * CollectorImpl
@@ -35,26 +32,63 @@ import java.util.Queue;
 public class CollectorImpl implements Collector
 {
 
-    private Queue<Event> events = new LinkedList<Event>();
+    private EventImpl head;
+    private EventImpl tail;
+    private EventImpl free;
 
     public CollectorImpl()
     {}
 
+    @Override
     public Event peek()
     {
-        return events.peek();
+        return head;
     }
 
+    @Override
     public void pop()
     {
-        events.poll();
+        if (head != null) {
+            EventImpl next = head.next;
+            head.next = free;
+            free = head;
+            head.clear();
+            head = next;
+        }
     }
 
-    public EventImpl put(Event.Type type)
+    public EventImpl put(Event.Type type, Object context)
     {
-        EventImpl event = new EventImpl(type);
-        events.add(event);
+        if (tail != null && tail.getType() == type &&
+            tail.getContext() == context) {
+            return null;
+        }
+
+        EventImpl event;
+        if (free == null) {
+            event = new EventImpl();
+        } else {
+            event = free;
+            free = free.next;
+            event.next = null;
+        }
+
+        event.init(type, context);
+
+        if (head == null) {
+            head = event;
+            tail = event;
+        } else {
+            tail.next = event;
+            tail = event;
+        }
+
         return event;
     }
 
+    @Override
+    public boolean more() {
+        return head != null && head.next != null;
+    }
+
 }
