@@ -29,6 +29,7 @@ import java.util.logging.Level;
 import java.util.logging.Logger;
 
 import org.apache.qpid.proton.amqp.Binary;
+import org.apache.qpid.proton.amqp.transport.EmptyFrame;
 import org.apache.qpid.proton.amqp.transport.FrameBody;
 import org.apache.qpid.proton.codec.ByteBufferDecoder;
 import org.apache.qpid.proton.codec.DecodeException;
@@ -40,6 +41,8 @@ class FrameParser implements TransportInput
 {
     private static final Logger TRACE_LOGGER = Logger.getLogger("proton.trace");
 
+    private static final ByteBuffer _emptyInputBuffer = newWriteableBuffer(0);
+
     private enum State
     {
         HEADER0,
@@ -62,12 +65,16 @@ class FrameParser implements TransportInput
 
     private final FrameHandler _frameHandler;
     private final ByteBufferDecoder _decoder;
+    private final int _inputBufferSize;
+    private final int _localMaxFrameSize;
 
-    private final ByteBuffer _inputBuffer;
+    private ByteBuffer _inputBuffer = null;
     private boolean _tail_closed = false;
 
     private State _state = State.HEADER0;
 
+    private long _framesInput = 0;
+
     /** the stated size of the current frame */
     private int _size;
 
@@ -82,16 +89,12 @@ class FrameParser implements TransportInput
      * We store the last result when processing input so that
      * we know not to process any more input if it was an error.
      */
-
-    FrameParser(FrameHandler frameHandler, ByteBufferDecoder decoder, int maxFrameSize)
+    FrameParser(FrameHandler frameHandler, ByteBufferDecoder decoder, int localMaxFrameSize)
     {
         _frameHandler = frameHandler;
         _decoder = decoder;
-        if (maxFrameSize > 0) {
-            _inputBuffer = newWriteableBuffer(maxFrameSize);
-        } else {
-            _inputBuffer = newWriteableBuffer(4*1024);
-        }
+        _localMaxFrameSize = localMaxFrameSize;
+        _inputBufferSize = _localMaxFrameSize > 0 ? _localMaxFrameSize : 4*1024;
     }
 
     private void input(ByteBuffer in) throws TransportException
@@ -290,6 +293,14 @@ class FrameParser implements TransportInput
                         break;
                     }
 
+                    if (_localMaxFrameSize > 0 && size > _localMaxFrameSize)
+                    {
+                        frameParsingError = new TransportException("specified frame size %d greater than maximum valid frame size %d",
+                                                                   size, _localMaxFrameSize);
+                        state = State.ERROR;
+                        break;
+                    }
+
                     if(in.remaining() < size-4)
                     {
                         _frameBuffer = ByteBuffer.allocate(size-4);
@@ -367,13 +378,16 @@ class FrameParser implements TransportInput
 
                     try
                     {
+                        _framesInput += 1;
+
+                        Binary payload = null;
+                        Object val = null;
+
                         if (frameBodySize > 0)
                         {
-
                             _decoder.setByteBuffer(in);
-                            Object val = _decoder.readObject();
-
-                            Binary payload;
+                            val = _decoder.readObject();
+                            _decoder.setByteBuffer(null);
 
                             if(in.hasRemaining())
                             {
@@ -385,41 +399,38 @@ class FrameParser implements TransportInput
                             {
                                 payload = null;
                             }
+                        }
+                        else
+                        {
+                            val = new EmptyFrame();
+                        }
 
-                            if(val instanceof FrameBody)
+                        if(val instanceof FrameBody)
+                        {
+                            FrameBody frameBody = (FrameBody) val;
+                            if(TRACE_LOGGER.isLoggable(Level.FINE))
                             {
-                                FrameBody frameBody = (FrameBody) val;
-                                if(TRACE_LOGGER.isLoggable(Level.FINE))
-                                {
-                                    TRACE_LOGGER.log(Level.FINE, "IN: CH["+channel+"] : " + frameBody + (payload == null ? "" : "[" + payload + "]"));
-                                }
-                                TransportFrame frame = new TransportFrame(channel, frameBody, payload);
-
-                                if(_frameHandler.isHandlingFrames())
-                                {
-                                    _tail_closed = _frameHandler.handleFrame(frame);
-                                }
-                                else
-                                {
-                                    transportAccepting = false;
-                                    _heldFrame = frame;
-                                }
+                                TRACE_LOGGER.log(Level.FINE, "IN: CH["+channel+"] : " + frameBody + (payload == null ? "" : "[" + payload + "]"));
+                            }
+                            TransportFrame frame = new TransportFrame(channel, frameBody, payload);
 
+                            if(_frameHandler.isHandlingFrames())
+                            {
+                                _tail_closed = _frameHandler.handleFrame(frame);
                             }
                             else
                             {
-                                throw new TransportException("Frameparser encountered a "
-                                        + (val == null? "null" : val.getClass())
-                                        + " which is not a " + FrameBody.class);
+                                transportAccepting = false;
+                                _heldFrame = frame;
                             }
                         }
                         else
                         {
-                            if(TRACE_LOGGER.isLoggable(Level.FINEST))
-                            {
-                                TRACE_LOGGER.finest("Ignored empty frame");
-                            }
+                            throw new TransportException("Frameparser encountered a "
+                                    + (val == null? "null" : val.getClass())
+                                    + " which is not a " + FrameBody.class);
                         }
+
                         reset();
                         in = oldIn;
                         oldIn = null;
@@ -447,7 +458,7 @@ class FrameParser implements TransportInput
                 state = State.ERROR;
                 frameParsingError = new TransportException("connection aborted");
             } else {
-                _frameHandler.closed();
+                _frameHandler.closed(null);
             }
         }
 
@@ -460,7 +471,7 @@ class FrameParser implements TransportInput
             if(frameParsingError != null)
             {
                 _parsingError = frameParsingError;
-                throw frameParsingError;
+                _frameHandler.closed(frameParsingError);
             }
             else
             {
@@ -475,35 +486,61 @@ class FrameParser implements TransportInput
         if (_tail_closed) {
             return Transport.END_OF_STREAM;
         } else {
-            return _inputBuffer.remaining();
+            if (_inputBuffer != null) {
+                return _inputBuffer.remaining();
+            } else {
+                return _inputBufferSize;
+            }
         }
     }
 
+    @Override
+    public int position() {
+        if (_tail_closed) {
+            return Transport.END_OF_STREAM;
+        }
+        return (_inputBuffer == null) ? 0 : _inputBuffer.position();
+    }
+
     @Override
     public ByteBuffer tail()
     {
         if (_tail_closed) {
-            if (_parsingError != null) {
-                throw new TransportException(_parsingError.getMessage());
-            } else {
-                throw new TransportException("tail closed");
-            }
+            throw new TransportException("tail closed");
         }
+
+        if (_inputBuffer == null) {
+            _inputBuffer = newWriteableBuffer(_inputBufferSize);
+        }
+
         return _inputBuffer;
     }
 
     @Override
     public void process() throws TransportException
     {
-        _inputBuffer.flip();
+        if (_inputBuffer != null)
+        {
+            _inputBuffer.flip();
 
-        try
-        {
-            input(_inputBuffer);
+            try
+            {
+                input(_inputBuffer);
+            }
+            finally
+            {
+                if (_inputBuffer.hasRemaining()) {
+                    _inputBuffer.compact();
+                } else if (_inputBuffer.capacity() > TransportImpl.BUFFER_RELEASE_THRESHOLD) {
+                    _inputBuffer = null;
+                } else {
+                    _inputBuffer.clear();
+                }
+            }
         }
-        finally
+        else
         {
-            _inputBuffer.compact();
+            input(_emptyInputBuffer);
         }
     }
 
@@ -542,4 +579,9 @@ class FrameParser implements TransportInput
         _size = 0;
         _state = State.SIZE_0;
     }
+
+    long getFramesInput()
+    {
+        return _framesInput;
+    }
 }
