@@ -573,8 +573,15 @@ public class MessageImpl implements ProtonJMessage
     @Override
     public int decode(byte[] data, int offset, int length)
     {
-        DecoderImpl decoder = tlsCodec.get().decoder;
         final ByteBuffer buffer = ByteBuffer.wrap(data, offset, length);
+        decode(buffer);
+
+        return length-buffer.remaining();
+    }
+
+    public void decode(ByteBuffer buffer)
+    {
+        DecoderImpl decoder = tlsCodec.get().decoder;
         decoder.setByteBuffer(buffer);
 
         _header = null;
@@ -680,9 +687,6 @@ public class MessageImpl implements ProtonJMessage
         }
 
         decoder.setByteBuffer(null);
-        
-        return length-buffer.remaining();
-
     }
 
     @Override
@@ -744,6 +748,9 @@ public class MessageImpl implements ProtonJMessage
         return length - buffer.remaining();
     }
 
+    /**
+     * @deprecated This method will be removed in a future release.
+     */
     @Override
     public void load(Object data)
     {
@@ -785,6 +792,9 @@ public class MessageImpl implements ProtonJMessage
 
     }
 
+    /**
+     * @deprecated This method will be removed in a future release.
+     */
     @Override
     public Object save()
     {
@@ -817,12 +827,18 @@ public class MessageImpl implements ProtonJMessage
         }
     }
 
+    /**
+     * @deprecated This method will be removed in a future release.
+     */
     @Override
     public String toAMQPFormat(Object value)
     {
         return _parser.encode(value);
     }
 
+    /**
+     * @deprecated This method will be removed in a future release.
+     */
     @Override
     public Object parseAMQPFormat(String value)
     {
@@ -831,12 +847,18 @@ public class MessageImpl implements ProtonJMessage
         return obj;
     }
 
+    /**
+     * @deprecated This method will be removed in a future release.
+     */
     @Override
     public void setMessageFormat(MessageFormat format)
     {
         _format = format;
     }
 
+    /**
+     * @deprecated This method will be removed in a future release.
+     */
     @Override
     public MessageFormat getMessageFormat()
     {
@@ -855,4 +877,24 @@ public class MessageImpl implements ProtonJMessage
         return MessageError.OK;
     }
 
+    public String toString()
+    {
+        StringBuilder sb = new StringBuilder();
+        sb.append("Message{");
+        if (_header != null) {
+            sb.append("header=");
+            sb.append(_header);
+        }
+        if (_properties != null) {
+            sb.append("properties=");
+            sb.append(_properties);
+        }
+        if (_body != null) {
+            sb.append("body=");
+            sb.append(_body);
+        }
+        sb.append("}");
+        return sb.toString();
+    }
+
 }
