@@ -1,3 +1,4 @@
+from __future__ import absolute_import
 #
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
@@ -17,13 +18,14 @@
 # under the License.
 #
 
-import os, common
+import os
+from . import common
 import random
 import string
 import subprocess
 
 from proton import *
-from common import Skipped, pump
+from .common import Skipped, pump
 
 
 def _testpath(file):
@@ -51,17 +53,24 @@ class SslTest(common.Test):
     class SslTestConnection(object):
         """ Represents a single SSL connection.
         """
-        def __init__(self, domain=None, session_details=None):
+        def __init__(self, domain=None, mode=Transport.CLIENT,
+                     session_details=None, conn_hostname=None,
+                     ssl_peername=None):
             if not common.isSSLPresent():
                 raise Skipped("No SSL libraries found.")
 
             self.ssl = None
             self.domain = domain
-            self.transport = Transport()
+            self.transport = Transport(mode)
             self.connection = Connection()
-            self.transport.bind(self.connection)
+            if conn_hostname:
+                self.connection.hostname = conn_hostname
             if domain:
                 self.ssl = SSL( self.transport, self.domain, session_details )
+                if ssl_peername:
+                    self.ssl.peer_hostname = ssl_peername
+            # bind last, after all configuration complete:
+            self.transport.bind(self.connection)
 
     def _pump(self, ssl_client, ssl_server, buffer_size=1024):
         pump(ssl_client.transport, ssl_server.transport, buffer_size)
@@ -73,6 +82,8 @@ class SslTest(common.Test):
         client.connection.open()
         server.connection.open()
         self._pump(client, server)
+        if client.transport.closed:
+            return
         assert client.ssl.protocol_name() is not None
         client.connection.close()
         server.connection.close()
@@ -82,7 +93,7 @@ class SslTest(common.Test):
         """ By default, both the server and the client support anonymous
         ciphers - they should connect without need for a certificate.
         """
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         # check that no SSL connection exists
@@ -111,7 +122,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.connection.open()
@@ -133,7 +144,7 @@ class SslTest(common.Test):
         self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                            self._testpath("server-private-key.pem"),
                                            "server-password")
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.connection.open()
@@ -154,7 +165,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.connection.open()
@@ -176,7 +187,7 @@ class SslTest(common.Test):
         self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.server_domain.set_peer_authentication( SSLDomain.VERIFY_PEER,
                                                     self._testpath("ca-certificate.pem") )
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
 
         # give the client a certificate, but let's not require server authentication
         self.client_domain.set_credentials(self._testpath("client-certificate.pem"),
@@ -204,7 +215,7 @@ class SslTest(common.Test):
         self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.server_domain.set_peer_authentication( SSLDomain.VERIFY_PEER,
                                                     self._testpath("ca-certificate.pem") )
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
 
         self.client_domain.set_credentials(self._testpath("bad-server-certificate.pem"),
                                            self._testpath("bad-server-private-key.pem"),
@@ -214,11 +225,11 @@ class SslTest(common.Test):
 
         client.connection.open()
         server.connection.open()
-        try:
-            self._pump( client, server )
-            assert False, "Server failed to reject bad certificate."
-        except TransportException, e:
-            pass
+        self._pump( client, server )
+        assert client.transport.closed
+        assert server.transport.closed
+        assert client.connection.state & Endpoint.REMOTE_UNINIT
+        assert server.connection.state & Endpoint.REMOTE_UNINIT
 
     def test_client_authentication_fail_no_cert(self):
         """ Ensure that the server will fail a client that does not provide a
@@ -232,18 +243,18 @@ class SslTest(common.Test):
         self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.server_domain.set_peer_authentication( SSLDomain.VERIFY_PEER,
                                                     self._testpath("ca-certificate.pem") )
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
 
         self.client_domain.set_peer_authentication( SSLDomain.ANONYMOUS_PEER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.connection.open()
         server.connection.open()
-        try:
-            self._pump( client, server )
-            assert False, "Server failed to reject bad certificate."
-        except TransportException, e:
-            pass
+        self._pump( client, server )
+        assert client.transport.closed
+        assert server.transport.closed
+        assert client.connection.state & Endpoint.REMOTE_UNINIT
+        assert server.connection.state & Endpoint.REMOTE_UNINIT
 
     def test_client_server_authentication(self):
         """ Require both client and server to mutually identify themselves.
@@ -261,7 +272,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.connection.open()
@@ -286,7 +297,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.connection.open()
@@ -309,16 +320,16 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.connection.open()
         server.connection.open()
-        try:
-            self._pump( client, server )
-            assert False, "Client failed to reject bad certificate."
-        except TransportException, e:
-            pass
+        self._pump( client, server )
+        assert client.transport.closed
+        assert server.transport.closed
+        assert client.connection.state & Endpoint.REMOTE_UNINIT
+        assert server.connection.state & Endpoint.REMOTE_UNINIT
 
         del server
         del client
@@ -327,7 +338,7 @@ class SslTest(common.Test):
         self.client_domain.set_peer_authentication( SSLDomain.ANONYMOUS_PEER )
 
         client = SslTest.SslTestConnection( self.client_domain )
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
 
         client.connection.open()
         server.connection.open()
@@ -348,7 +359,7 @@ class SslTest(common.Test):
                                                     self._testpath("ca-certificate.pem") )
         # allow unsecured clients on this connection
         self.server_domain.allow_unsecured_client()
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
 
         # non-ssl connection
         client = SslTest.SslTestConnection()
@@ -383,7 +394,7 @@ class SslTest(common.Test):
 
         # client uses ssl. Server should detect this.
         client = SslTest.SslTestConnection( self.client_domain )
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
 
         client.connection.open()
         server.connection.open()
@@ -402,18 +413,18 @@ class SslTest(common.Test):
                                            "server-password")
         self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.server_domain.set_peer_authentication( SSLDomain.ANONYMOUS_PEER )
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
 
         # non-ssl connection
         client = SslTest.SslTestConnection()
 
         client.connection.open()
         server.connection.open()
-        try:
-            self._pump( client, server )
-            assert False, "Server did not reject client as expected."
-        except TransportException:
-            pass
+        self._pump( client, server )
+        assert client.transport.closed
+        assert server.transport.closed
+        assert client.connection.state & Endpoint.REMOTE_UNINIT
+        assert server.connection.state & Endpoint.REMOTE_UNINIT
 
     def test_session_resume(self):
         """ Test resume of client session.
@@ -429,8 +440,8 @@ class SslTest(common.Test):
         # details will be used in initial and subsequent connections to allow session to be resumed
         initial_session_details = SSLSessionDetails("my-session-id")
 
-        server = SslTest.SslTestConnection( self.server_domain )
-        client = SslTest.SslTestConnection( self.client_domain, initial_session_details )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
+        client = SslTest.SslTestConnection( self.client_domain, session_details=initial_session_details )
 
         # bring up the connection and store its state
         client.connection.open()
@@ -448,9 +459,9 @@ class SslTest(common.Test):
         del server
 
         # now create a new set of connections, use last session id
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         # provide the details of the last session, allowing it to be resumed
-        client = SslTest.SslTestConnection( self.client_domain, initial_session_details )
+        client = SslTest.SslTestConnection( self.client_domain, session_details=initial_session_details )
 
         #client.transport.trace(Transport.TRACE_DRV)
         #server.transport.trace(Transport.TRACE_DRV)
@@ -475,8 +486,8 @@ class SslTest(common.Test):
         del client
         del server
 
-        server = SslTest.SslTestConnection( self.server_domain )
-        client = SslTest.SslTestConnection( self.client_domain, SSLSessionDetails("some-other-session-id") )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
+        client = SslTest.SslTestConnection( self.client_domain, session_details=SSLSessionDetails("some-other-session-id") )
 
         client.connection.open()
         server.connection.open()
@@ -507,7 +518,7 @@ class SslTest(common.Test):
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER )
 
         max_count = 100
-        sessions = [(SslTest.SslTestConnection( self.server_domain ),
+        sessions = [(SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER ),
                      SslTest.SslTestConnection( self.client_domain )) for x in
                     range(max_count)]
         for s in sessions:
@@ -541,7 +552,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "a1.good.server.domain.com"
@@ -559,15 +570,15 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "A1.Good.Server.domain.comX"
-        try:
-            self._do_handshake( client, server )
-            assert False, "Expected connection to fail due to hostname mismatch"
-        except TransportException:
-            pass
+        self._do_handshake( client, server )
+        assert client.transport.closed
+        assert server.transport.closed
+        assert client.connection.state & Endpoint.REMOTE_UNINIT
+        assert server.connection.state & Endpoint.REMOTE_UNINIT
         del server
         del client
         self.teardown()
@@ -587,7 +598,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "alternate.Name.one.com"
@@ -604,7 +615,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "ANOTHER.NAME.COM"
@@ -621,7 +632,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "SOME.PREfix.domain.COM"
@@ -638,7 +649,7 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "FOO.PREfixZZZ.domain.com"
@@ -655,15 +666,15 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "FOO.PREfi.domain.com"
-        try:
-            self._do_handshake( client, server )
-            assert False, "Expected connection to fail due to hostname mismatch"
-        except TransportException:
-            pass
+        self._do_handshake( client, server )
+        assert client.transport.closed
+        assert server.transport.closed
+        assert client.connection.state & Endpoint.REMOTE_UNINIT
+        assert server.connection.state & Endpoint.REMOTE_UNINIT
         del server
         del client
         self.teardown()
@@ -676,17 +687,51 @@ class SslTest(common.Test):
         self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
         self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
 
-        server = SslTest.SslTestConnection( self.server_domain )
+        server = SslTest.SslTestConnection( self.server_domain, mode=Transport.SERVER )
         client = SslTest.SslTestConnection( self.client_domain )
 
         client.ssl.peer_hostname = "PREfix.domain.COM"
-        try:
-            self._do_handshake( client, server )
-            assert False, "Expected connection to fail due to hostname mismatch"
-        except TransportException:
-            pass
+        self._do_handshake( client, server )
+        assert client.transport.closed
+        assert server.transport.closed
+        assert client.connection.state & Endpoint.REMOTE_UNINIT
+        assert server.connection.state & Endpoint.REMOTE_UNINIT
         self.teardown()
 
+        # Pass: ensure that the user can give an alternate name that overrides
+        # the connection's configured hostname
+        self.setup()
+        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
+                                    self._testpath("server-wc-private-key.pem"),
+                                    "server-password")
+        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
+        self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
+
+        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
+        client = SslTest.SslTestConnection(self.client_domain,
+                                           conn_hostname="This.Name.Does.not.Match",
+                                           ssl_peername="alternate.name.one.com")
+        self._do_handshake(client, server)
+        del client
+        del server
+        self.teardown()
+
+        # Pass: ensure that the hostname supplied by the connection is used if
+        # none has been specified for the SSL instanace
+        self.setup()
+        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
+                                    self._testpath("server-private-key.pem"),
+                                    "server-password")
+        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
+        self.client_domain.set_peer_authentication( SSLDomain.VERIFY_PEER_NAME )
+
+        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
+        client = SslTest.SslTestConnection(self.client_domain,
+                                           conn_hostname="a1.good.server.domain.com")
+        self._do_handshake(client, server)
+        del client
+        del server
+        self.teardown()
 
     def test_defaults_messenger_app(self):
         """ Test an SSL connection using the Messenger apps (no certificates)
@@ -789,7 +834,7 @@ class SslTest(common.Test):
         try:
             ssl3 = SSL(transport, self.server_domain)
             assert False, "Expected error did not occur!"
-        except SSLException, e:
+        except SSLException:
             pass
 
 class MessengerSSLTests(common.Test):
@@ -832,7 +877,8 @@ class MessengerSSLTests(common.Test):
             self.server.subscribe("amqps://~0.0.0.0:12345")
             if exception is not None:
                 assert False, "expected failure did not occur"
-        except MessengerException, e:
+        except MessengerException:
+            e = sys.exc_info()[1]
             if exception:
                 assert exception in str(e), str(e)
             else:
